// BASEDISP.CPP
//
// Standard base dispatch interface from which to derive project-specific
// classes. This interface takes care of the boring grunt work required
// for all Visual Basic classes.
//

#include "stdhdr.h"
#include "BaseDisp.h"
#include "srvdata.h"
#include "sieve.h"

// Support function for deriving
void CBaseIDispatch::Derive(CBaseIDispatch FAR * pDest, CBaseIDispatch FAR * pSource)
{
	if (pSource) {
		CBaseIDispatch &refDest = *pDest;
		CBaseIDispatch &refSrc = *pSource;
		if (refSrc.m_pSupportErrorInfo) {
			refDest.m_pSupportErrorInfo = refSrc.m_pSupportErrorInfo;
			refDest.m_pSupportErrorInfo->AddRef();
		}
	}
}

// CBaseIDispatch: base constructor to support IDispatch
CBaseIDispatch::CBaseIDispatch()
{
	m_refs = 0;
	m_pIID = &IID_NULL;
	m_ptinfo = NULL;
	m_pSupportErrorInfo = NULL; 
	g_cObj++;
}

// CBaseIDispatch: base destructor to support IDispatch
CBaseIDispatch::~CBaseIDispatch()
{
	if (m_ptinfo) {
		m_ptinfo->Release();
	}
	if (m_pSupportErrorInfo) {
		m_pSupportErrorInfo->Release();
	}
	g_cObj--;
}

// Standard IUnknown methods
STDMETHODIMP
CBaseIDispatch::QueryInterface(THIS_ REFIID riid, LPVOID FAR * ppvObj)
{
	// Return self if requested with IUnknown, IDispatch, or self
	if (IIDOK(IUnknown) || IIDOK(IDispatch) || IsEqualIID(riid, *m_pIID)) {
		*ppvObj = this;
	// Let error object return itself
	} else if (m_pSupportErrorInfo && IIDOK(ISupportErrorInfo)) {
		return m_pSupportErrorInfo->QueryInterface(riid, ppvObj);
	// Provide class information from type library
	} else if (IIDOK(IProvideClassInfo)) {
		HRESULT hr;
		if SUCCEEDED(hr = LoadTypeInfo()) {
			return CProvideClassInfo::Create(m_ptinfo,
									  (CProvideClassInfo FAR * FAR *) ppvObj);
		} else {
			return hr;
		}
	} else {
		*ppvObj = NULL;
		return ResultFromScode(E_NOINTERFACE);
	}
	AddRef();
	return NOERROR;
}

STDMETHODIMP_(ULONG)
CBaseIDispatch::AddRef(void)
{
	return ++m_refs;
}

STDMETHODIMP_(ULONG)
CBaseIDispatch::Release(void)
{
	if (--m_refs == 0) {
		delete this;                                
		return 0;
	}
	QueryUnregisterObject();
	return m_refs;
}

// Standard IDispatch methods
STDMETHODIMP
CBaseIDispatch::GetTypeInfoCount(UINT FAR * pctinfo)
{
    *pctinfo = 1;
    return NOERROR;
}

STDMETHODIMP
CBaseIDispatch::GetTypeInfo(UINT itinfo, LCID lcid,
							ITypeInfo FAR * FAR * pptinfo)
{

	HRESULT hr;
	if SUCCEEDED(hr = LoadTypeInfo()) {
		m_ptinfo->AddRef();
		*pptinfo = m_ptinfo;
	} else {
		*pptinfo = NULL;
	}
    return hr;
}

STDMETHODIMP
CBaseIDispatch::GetIDsOfNames(REFIID riid, OLECHAR FAR * FAR * rgszNames,
							  UINT cNames, LCID lcid, DISPID FAR * rgdispid)
{
    HRESULT hr;
	if (riid != IID_NULL) {
        hr = ResultFromScode(DISP_E_UNKNOWNINTERFACE);
	} else if SUCCEEDED(hr = LoadTypeInfo()) {
		// Use standard implementation
		hr = DispGetIDsOfNames(m_ptinfo, rgszNames, cNames, rgdispid);
	}
	return hr;
}

STDMETHODIMP
CBaseIDispatch::Invoke(DISPID dispidMember, REFIID riid, LCID lcid,
					   WORD wFlags, DISPPARAMS FAR * pdispparams,
					   VARIANT FAR * pvarResult, EXCEPINFO FAR * pexcepinfo,
					   UINT FAR * puArgErr)
{
    HRESULT hr;
	if (riid != IID_NULL) {
		hr = ResultFromScode(DISP_E_UNKNOWNINTERFACE);
	} else if SUCCEEDED(hr = LoadTypeInfo()) {
		// Use standard implementation
		hr = DispInvoke(this, m_ptinfo, dispidMember, wFlags,
						pdispparams, pvarResult, pexcepinfo, puArgErr);
	}
	return hr;        
}

HRESULT CBaseIDispatch::LoadTypeInfo()
{
	HRESULT hr;
	if (m_ptinfo) {
		hr = NOERROR;
	} else {
		hr = LoadResourceTypeLib(LIBID_SieveServerCpp, *m_pIID, &m_ptinfo);
	}
	return hr;
}

// RegisterObject and QueryUnregisterObject correctly register or unregister
// this object in RunningObjectTable
void CBaseIDispatch::RegisterObject(const IID FAR * lpclsid)
{
	m_dwRegister = NULL;
	m_RegisterCount = m_refs;
	if (S_OK != RegisterActiveObject(this, *lpclsid,
									 ACTIVEOBJECT_WEAK, &m_dwRegister)) {
		m_dwRegister = NULL;
	}
	m_RegisterCount = m_refs - m_RegisterCount;
	return;
}

void CBaseIDispatch::QueryUnregisterObject(void)
{
	if (m_dwRegister && m_refs == m_RegisterCount) {
		RevokeActiveObject(m_dwRegister, NULL);
		m_dwRegister = NULL;
	}
}

// CProvideClassInfo implementation
HRESULT CProvideClassInfo::Create(ITypeInfo FAR * ptinfo,
								  CProvideClassInfo FAR * FAR * retVal)
{
	if (NULL == (*retVal = new CProvideClassInfo())) {
		return ResultFromScode(E_OUTOFMEMORY);
	} else {
		(*retVal)->m_ptinfo = ptinfo;
		ptinfo->AddRef();
		(*retVal)->AddRef();
		return NOERROR;
	}
}

// Constructor
CProvideClassInfo::CProvideClassInfo()
{
	m_refs = 0;
	g_cObj++;
}

// Destructor
CProvideClassInfo::~CProvideClassInfo()
{
	if (m_ptinfo) {
		m_ptinfo->Release();
	}
	g_cObj--;
}

// Standard IUnknown members
STDMETHODIMP
CProvideClassInfo::QueryInterface(THIS_ REFIID riid, LPVOID FAR * ppvObj)
{
	if (IIDOK(IProvideClassInfo) || IIDOK(IUnknown)) {
		*ppvObj = this;
	} else {
		*ppvObj = NULL;
		return ResultFromScode(E_NOINTERFACE);
	}
	AddRef();
	return NOERROR;
}

STDMETHODIMP_(ULONG)
CProvideClassInfo::AddRef(void)
{
	return ++m_refs;
}

STDMETHODIMP_(ULONG)
CProvideClassInfo::Release(void)
{
	if (--m_refs == 0) {
		delete this;                                
		return 0;
	}
	return m_refs;
}

STDMETHODIMP
CProvideClassInfo::GetClassInfo(THIS_ LPTYPEINFO FAR * ppTI)
{
	// Return information from type library
	if (m_ptinfo) {
		m_ptinfo->AddRef();
		*ppTI = m_ptinfo;
		return NOERROR;
	} else {
		*ppTI = NULL;
		return ResultFromScode(E_FAIL);
	}
}
